#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include "omf.h"


int OMFProcess (FILE *f, OMFProcessor *proc)
{
  /* Improve I/O performance by buffering the input */
  const int file_buffer_size = 10240;
  char *file_buffer = (char *) malloc (file_buffer_size);
  if (file_buffer)
    setvbuf (f, file_buffer, _IOFBF, file_buffer_size);

  /* Make an internal buffer for holding temp data */
  const int omf_record_buffer_size = 256 * 256 + 10;
  byte *omf_record_buffer = (byte *) malloc (omf_record_buffer_size);
  /* Count of data already read into buffer */
  int data_fill_size = 0;
  proc->parse_stop = 0;
  for (;;)
    {
      if (proc->omf_error)
	return 0;
      if (proc->parse_stop)
	return 1;
      /* read an omf record */
      const int read_count =
	fread (omf_record_buffer + data_fill_size, 1,
	       omf_record_buffer_size - data_fill_size, f);
      const int valid_data_count = read_count + data_fill_size;
      
      /* detect eof */
      if (valid_data_count == 0)
	break;
      
      /* No omf record is smaller than 4 bytes */
      if (valid_data_count < 4)
	{
	  proc->format_error ("Premature end of file(1)");
	  return 0;
	}

      /* Get record type and record length */
      const byte record_type = omf_record_buffer[0];
      const unsigned short record_len =
	extract_omf_short (omf_record_buffer + 1);

      /* Validate record length */
      if (record_len < 1)
	{
	  proc->format_error ("Premature end of file(2)");
	  return 0 ;
	}

      /* Validate that the record is completely read */
      if (valid_data_count < record_len + 3)
	{
	  proc->format_error ("Premature end of file(3)");
	  return 0;
	}

      /* process the raw portion of the record */
      byte * const raw_data = &omf_record_buffer[3];
      const unsigned short raw_data_length = record_len - 1;

      proc->raw_record (record_type, raw_data, raw_data_length);

      /* Move the unused data to the front of the record */
      const unsigned short used_data = record_len + 3;
      if (used_data < valid_data_count)
	{
	  data_fill_size = valid_data_count - used_data;
	  memmove (omf_record_buffer, omf_record_buffer + used_data,
		   data_fill_size);
	}
      else
	data_fill_size = 0;
    }
  return 1;
}

OMFProcessor::OMFProcessor ()
: index_alloc_size (0),
  index_size (0),
  index (0),
  parse_stop (0),
  omf_error (0)
{
}

OMFProcessor::~OMFProcessor ()
{
  int i;
  for (i = 0; i < index_size; i++)
    free (index[i]);
  if (index)
    free (index);
}

void OMFProcessor::append_index_name (const PString *name)
{
  if (index_alloc_size == 0)
    {
      index_alloc_size = 256;
      index = (PString **) malloc (index_alloc_size * sizeof (index[0]));
    }
  else if (index_size >= index_alloc_size)
    {
      index_alloc_size += 256;
      index = (PString **) realloc (index,
				   index_alloc_size * sizeof (index[0]));
    }
  PString *copy = Pstrdup (name);
  if (index == 0 || copy == 0)
    {
      fprintf (stderr, "Out of memory\n");
      exit (1);
    }
  index[index_size++] = copy;
}

const PString *OMFProcessor::lookup_index_name (unsigned short idx)
{
  idx--;
  if (idx >= index_size)
    return 0;
  else
    return index[idx];
}

virtual void OMFProcessor::raw_record (byte type,
				       byte *data, unsigned short len)
{
  switch (type)
    {
    case 0x80:
      raw_theadr (data, len);
      break;
    case 0x82:
      raw_lheadr (data, len);
      break;
    case 0x88:
      raw_coment (data, len);
      break;
    case 0x8A:
      raw_modend (data, len);
      break;
    case 0x8B:
      raw_modend32 (data, len);
      break;
    case 0x8C:
      raw_extdef (data, len);
      break;
    case 0x90:
      raw_pubdef (data, len);
      break;
    case 0x91:
      raw_pubdef32 (data, len);
      break;
    case 0x94:
      raw_linnum (data, len);
      break;
    case 0x95:
      raw_linnum32 (data, len);
      break;
    case 0x96:
      raw_lnames (data, len);
      break;
    case 0x98:
      raw_segdef (data, len);
      break;
    case 0x99:
      raw_segdef32 (data, len);
      break;
    case 0x9A:
      raw_grpdef (data, len);
      break;
    case 0x9C:
      raw_fixupp (data, len);
      break;
    case 0x9D:
      raw_fixupp32 (data, len);
      break;
    case 0xA0:
      raw_ledata (data, len);
      break;
    case 0xA1:
      raw_ledata32 (data, len);
      break;
    case 0xA2:
      raw_lidata (data, len);
      break;
    case 0xA3:
      raw_lidata32 (data, len);
      break;
    case 0xB0:
      raw_comdef (data, len);
      break;
    case 0xB2:
      raw_bakpat (data, len);
      break;
    case 0xB3:
      raw_bakpat32 (data, len);
      break;
    case 0xB4:
    case 0xB5:
      raw_lextdef (data, len);
      break;
    case 0xB6:
      raw_lpubdef (data, len);
      break;
    case 0xB7:
      raw_lpubdef32 (data, len);
      break;
    case 0xB8:
      raw_lcomdef (data, len);
      break;
    case 0xBC:
      raw_cextdef (data, len);
      break;
    case 0xC2:
      raw_comdat (data, len);
      break;
    case 0xC3:
      raw_comdat32 (data, len);
      break;
    case 0xC4:
      raw_linsym (data, len);
      break;
    case 0xC5:
      raw_linsym32 (data, len);
      break;
    case 0xC6:
      raw_alias (data, len);
      break;
    case 0xC8:
      raw_nbkpat (data, len);
      break;
    case 0xC9:
      raw_nbkpat32 (data, len);
      break;
    case 0xCA:
      raw_llnames (data, len);
      break;
    case 0xCC:
      raw_vernum (data, len);
      break;
    case 0xCE:
      raw_vendext (data, len);
      break;
    case 0xF0:
      raw_library_header_record (data, len);
      break;
    case 0xF1:
      raw_library_end_record (data, len);
      break;
    default:
      unknown_record (type, data, len);
      break;
    };
}

virtual void OMFProcessor::unknown_record (byte type, byte *data,
					   unsigned short len)
{
}

void OMFProcessor::format_error (const char *err, ...)
{
  omf_error = 1;
  va_list va;
  va_start (va, err);
  va_format_error (err, va);
  va_end (va);
}

void OMFProcessor::va_format_error (const char *, va_list)
{
}

virtual void OMFProcessor::raw_theadr (byte *data, unsigned short len)
{
  if (data[0] + 1 != len)
    format_error ("THEADR contains invalid counted string");
  else
    theadr ((PString *) data);
}

void OMFProcessor::theadr (PString *)
{
}


virtual void OMFProcessor::raw_lheadr (byte *data, unsigned short len)
{
}

virtual void OMFProcessor::raw_coment (byte *data, unsigned short len)
{
  if (len < 2)
    format_error ("COMENT record is too short");
  else
    {
//      int nopurge = !!(data[0] & 0x80);
//      int nolist = !!(data[0] & 0x40);
      coment (!!(data[0] & 0x80), !!(data[0] & 0x40), data[1],
	      data + 2, len - 2);
    }
}

void OMFProcessor::coment (int nopurge, int nolist, int coment_class,
			   byte *data, unsigned short len)
{
  switch (coment_class)
    {
    case 0x00:
      if (len == 0)
	format_error ("COMENT TRANSLATOR record contains no string");
      else
	coment_translator (data, len);
      break;
    case 0x9F:
      if (len == 0)
	format_error ("COMENT Default Library Record contains no string");
      else
	coment_default_library (data, len);
      break;
    case 0xA2:
      coment_end_of_link_pass_one ();
      break;
    default:
      coment_other (coment_class, data, len);
      break;
    }
}

void OMFProcessor::coment_translator (byte *name, unsigned len)
{
}

void OMFProcessor::coment_default_library (byte *name, unsigned len)
{
}

void OMFProcessor::coment_end_of_link_pass_one ()
{
}

void OMFProcessor::coment_other (int coment_class,
				 byte *data, unsigned short len)
{
}

virtual void OMFProcessor::raw_modend (byte *data, unsigned short len)
{
}

virtual void OMFProcessor::raw_modend32 (byte *data, unsigned short len)
{
}

virtual void OMFProcessor::raw_extdef (byte *data, unsigned short len)
{
  if (len == 0
      || len < data[0] + 1)
    {
      format_error ("EXTDEF record too short");
      return;
    }
  PString *s = (PString *) data;
  len -= data[0] + 1;
  data += data[0] + 1;
  if (len == 0
      || (len == 1 && !omf_index_is_one_byte (data)))
    {
      format_error ("EXTDEF record too short");
      return;
    }
  int typelen;
  unsigned short type = extract_omf_index (data, &typelen);
  if (len != typelen)
    {
      format_error ("EXTDEF record too long");
      return;
    }
  extdef (s, type);
}

void OMFProcessor::extdef (PString *name, unsigned short type)
{
}

virtual void OMFProcessor::raw_pubdef (byte *data, unsigned short len)
{
}

virtual void OMFProcessor::raw_pubdef32 (byte *data, unsigned short len)
{
  /* Parse base group */
  if (len < 2)
    {
      format_error ("PUBDEF32 record too short(1)");
      return;
    }
  int datalen;
  unsigned short base_group = extract_omf_index (data, &datalen);
  data += datalen;
  len -= datalen;

  /* Parse base segment */
  if (len < 2)
    {
      format_error ("PUBDEF32 record too short(2)");
      return;
    }
  unsigned short base_segment = extract_omf_index (data, &datalen);
  data += datalen;
  len -= datalen;

  /* If group index is zero, skip the unused base frame field */
  if (base_segment == 0)
    {
      data += 2;
      len -= 2;
    }

  /* Parse the string */
  if (len == 0
      || len < data[0] + 1
      || data[0] == 0)
    {
      format_error ("PUBDEF32 record too short(3)");
      return;
    }
  PString *public_name = (PString *) data;
  data += public_name->count + 1;
  len -= public_name->count + 1;

  /* Get the public offset */
  if (len < 4)
    {
      format_error ("PUBDEF32 record too short(4)");
      return;
    }
  unsigned long public_offset = extract_omf_long (data);
  data += 4;
  len -= 4;

  /* Get the type */
  if (len == 0
      || (len == 1 && !omf_index_is_one_byte (data)))
    {
      format_error ("PUBDEF32 record too short(5)");
      return;
    }
  unsigned short type = extract_omf_index (data, &datalen);
  if (datalen != len)
    {
      format_error ("PUBDEF32 record too long");
      return;
    }
  
  pubdef32 (base_group, base_segment, public_name, public_offset, type);
}

void OMFProcessor::pubdef32 (unsigned short group, unsigned short segment,
			     PString *name, unsigned long offset,
			     unsigned short type)
{
}

virtual void OMFProcessor::raw_linnum (byte *data, unsigned short len)
{
}

virtual void OMFProcessor::raw_linnum32 (byte *data, unsigned short len)
{
}

virtual void OMFProcessor::raw_lnames (byte *data, unsigned short len)
{
  while (len != 0)
    if (data[0]+1 > len)
      {
	format_error ("LNAME record contains invalid data");
	return;
      }
    else
      {
	lname ((PString *) data);
	len -= data[0] + 1;
	data += data[0] + 1;
      }
}

void OMFProcessor::lname (PString *name)
{
  append_index_name (name);
}

virtual void OMFProcessor::raw_segdef (byte *data, unsigned short len)
{
}

virtual void OMFProcessor::raw_segdef32 (byte *data, unsigned short len)
{
  /* Parse the attributes byte */
  if (len == 0)
    {
      format_error ("Empty SEGDEF32 record");
      return;
    }
  const byte alignment = (data[0] >> 5);
  const byte combination = ((data[0] >> 2) & 7);
  const byte big = !!(data[0] & 2);
  const byte bit32 = !!(data[0] & 1);
  len--;
  data++;

  /* Ignore the frame number and offset for absolute segments */
  if (alignment == 0)
    if (len < 3)
      {
	format_error ("SEGDEF32 too short");
	return;
      }
    else
      {
	len -= 3;
	data += 3;
      }

  /* Parse segment length */
  if (len < 4)
    {
      format_error ("SEGDEF32 too short");
      return;
    }
  const unsigned long segment_length = extract_omf_long (data);
  data += 4;
  len -= 4;

  /* Extract indexes for segment name, class name, overlay name */
  unsigned short indexes[3];
  int i;
  for (i = 0; i < 3; i++)
    if (len < 1
	|| (len == 1 && !omf_index_is_one_byte (data)))
      {
	format_error ("SEGDEF32 too short");
	return;
      }
    else
      {
	int index_length;
	indexes[i] = extract_omf_index (data, &index_length);
	data += index_length;
	len -= index_length;
      }

  if (len != 0)
    {
      format_error ("SEGDEF32 record too long");
    }

  const PString *segment_name = lookup_index_name (indexes[0]);
  if (segment_name == 0)
    {
      format_error ("SEGDEF32 contains invalid segment string index: %d",
		    indexes[0]);
      return;
    }
  const PString *class_name = lookup_index_name (indexes[1]);
  if (class_name == 0)
    {
      format_error ("SEGDEF32 contains invalid class string index: %d",
		    indexes[1]);
      return;
    }
  const PString *overlay_name = lookup_index_name (indexes[2]);
  if (overlay_name == 0)
    {
      format_error ("SEGDEF32 contains invalid overlay string index: %d",
		    indexes[2]);
      return;
    }

  segdef32 (alignment, combination, big, bit32, segment_length,
	    segment_name, class_name, overlay_name);
}

void OMFProcessor::segdef32 (byte alignment, byte combination, byte big,
			     byte bit32, unsigned long segment_length,
			     const PString *segment_name,
			     const PString *class_name,
			     const PString *overlay_name)
{
}

virtual void OMFProcessor::raw_grpdef (byte *data, unsigned short len)
{
}

virtual void OMFProcessor::raw_fixupp (byte *data, unsigned short len)
{
}

virtual void OMFProcessor::raw_fixupp32 (byte *data, unsigned short len)
{
}

virtual void OMFProcessor::raw_ledata (byte *data, unsigned short len)
{
}

virtual void OMFProcessor::raw_ledata32 (byte *data, unsigned short len)
{
}

virtual void OMFProcessor::raw_lidata (byte *data, unsigned short len)
{
}

virtual void OMFProcessor::raw_lidata32 (byte *data, unsigned short len)
{
}

void OMFProcessor::raw_comdef (byte *data, unsigned short len)
{
  while (len > 0)
    {
      /* Parse name */
      if (len < data[0] + 1)
	{
	  format_error ("COMDEF record too short");
	  return;
	}
      PString *name = (PString *) data;
      len -= name->count + 1;
      data += name->count + 1;
      
      /* Parse type index */
      if (len < 2)
	{
	  format_error ("COMDEF record too short");
	  return;
	}
      int datalen;
      unsigned short type_index = extract_omf_index (data, &datalen);
      len -= datalen;
      data += datalen;
      
      /* Skip data type */
      if (len < 1)
	{
	  format_error ("COMDEF record too short");
	  return;
	}
      len -= 1;
      data += 1;

      /* Parse communal length */
      if (len < 1)
	{
	  format_error ("COMDEF record too short");
	  return;
	}
      unsigned long communal_length;
      switch (data[0])
	{
	case 0x81: /* 16 bits follow */
	  if (len < 3)
	    {
	      format_error ("COMDEF record too short");
	      return;
	    }
	  communal_length = extract_omf_short (data + 1);
	  data += 3;
	  len -= 3;
	  break;
	case 0x84: /* 24 bits follow */
	  if (len < 4)
	    {
	      format_error ("COMDEF record too short");
	      return;
	    }
	  communal_length = extract_omf_short (data + 1);
	  communal_length += (((long) data[3]) * 256 * 256);
	  data += 4;
	  len -= 4;
	  break;
	case 0x88: /* 32 bits follow */
	  if (len < 5)
	    {
	      format_error ("COMDEF record too short");
	      return;
	    }
	  communal_length = extract_omf_long (data + 1);
	  data += 5;
	  len -= 5;
	  break;
	default:
	  if (data[0] > 128)
	    {
	      format_error ("Unknown COMDEF length code");
	      return;
	    }
	  communal_length = data[0];
	  data += 1;
	  len -= 1;
	  break;
	}

      comdef (name, type_index, communal_length);
    }
}

void OMFProcessor::comdef (PString *name, unsigned short type,
			   unsigned long length)
{
}

virtual void OMFProcessor::raw_bakpat (byte *data, unsigned short len)
{
}

virtual void OMFProcessor::raw_bakpat32 (byte *data, unsigned short len)
{
}

virtual void OMFProcessor::raw_lextdef (byte *data, unsigned short len)
{
  if (len == 0
      || len < data[0] + 1)
    {
      format_error ("LEXTDEF record too short");
      return;
    }
  PString *s = (PString *) data;
  len -= data[0] + 1;
  data += data[0] + 1;
  if (len == 0
      || (len == 1 && !omf_index_is_one_byte (data)))
    {
      format_error ("LEXTDEF record too short");
      return;
    }
  int typelen;
  unsigned short type = extract_omf_index (data, &typelen);
  if (len != typelen)
    {
      format_error ("LEXTDEF record too long");
      return;
    }
  lextdef (s, type);
}

void OMFProcessor::lextdef (PString *name, unsigned short type)
{
}

virtual void OMFProcessor::raw_lpubdef (byte *data, unsigned short len)
{
}

virtual void OMFProcessor::raw_lpubdef32 (byte *data, unsigned short len)
{
  /* Parse base group */
  if (len < 2)
    {
      format_error ("LPUBDEF32 record too short(1)");
      return;
    }
  int datalen;
  unsigned short base_group = extract_omf_index (data, &datalen);
  data += datalen;
  len -= datalen;

  /* Parse base segment */
  if (len < 2)
    {
      format_error ("LPUBDEF32 record too short(2)");
      return;
    }
  unsigned short base_segment = extract_omf_index (data, &datalen);
  data += datalen;
  len -= datalen;

  /* If group index is zero, skip the unused base frame field */
  if (base_segment == 0)
    {
      data += 2;
      len -= 2;
    }

  /* Parse the string */
  if (len == 0
      || len < data[0] + 1
      || data[0] == 0)
    {
      format_error ("LPUBDEF32 record too short(3)");
      return;
    }
  PString *public_name = (PString *) data;
  data += public_name->count + 1;
  len -= public_name->count + 1;

  /* Get the public offset */
  if (len < 4)
    {
      format_error ("LPUBDEF32 record too short(4)");
      return;
    }
  unsigned long public_offset = extract_omf_long (data);
  data += 4;
  len -= 4;

  /* Get the type */
  if (len == 0
      || (len == 1 && !omf_index_is_one_byte (data)))
    {
      format_error ("LPUBDEF32 record too short(5)");
      return;
    }
  unsigned short type = extract_omf_index (data, &datalen);
  if (datalen != len)
    {
      format_error ("LPUBDEF32 record too long");
      return;
    }
  
  lpubdef32 (base_group, base_segment, public_name, public_offset, type);
}

void  OMFProcessor::lpubdef32 (unsigned short group, unsigned short segment,
			       PString *name, unsigned long offset,
			       unsigned short type)
{
}

virtual void OMFProcessor::raw_lcomdef (byte *data, unsigned short len)
{
  while (len > 0)
    {
      /* Parse name */
      if (len < data[0] + 1)
	{
	  format_error ("LCOMDEF record too short");
	  return;
	}
      PString *name = (PString *) data;
      len -= name->count + 1;
      data += name->count + 1;
      
      /* Parse type index */
      if (len < 2)
	{
	  format_error ("LCOMDEF record too short");
	  return;
	}
      int datalen;
      unsigned short type_index = extract_omf_index (data, &datalen);
      len -= datalen;
      data += datalen;
      
      /* Skip data type */
      if (len < 1)
	{
	  format_error ("LCOMDEF record too short");
	  return;
	}
      len -= 1;
      data += 1;

      /* Parse communal length */
      if (len < 1)
	{
	  format_error ("LCOMDEF record too short");
	  return;
	}
      unsigned long communal_length;
      switch (data[0])
	{
	case 0x81: /* 16 bits follow */
	  if (len < 3)
	    {
	      format_error ("LCOMDEF record too short");
	      return;
	    }
	  communal_length = extract_omf_short (data + 1);
	  data += 3;
	  len -= 3;
	  break;
	case 0x84: /* 24 bits follow */
	  if (len < 4)
	    {
	      format_error ("LCOMDEF record too short");
	      return;
	    }
	  communal_length = extract_omf_short (data + 1);
	  communal_length += (((long) data[3]) * 256 * 256);
	  data += 4;
	  len -= 4;
	  break;
	case 0x88: /* 32 bits follow */
	  if (len < 5)
	    {
	      format_error ("LCOMDEF record too short");
	      return;
	    }
	  communal_length = extract_omf_long (data + 1);
	  data += 5;
	  len -= 5;
	  break;
	default:
	  if (data[0] > 128)
	    {
	      format_error ("Unknown LCOMDEF length code");
	      return;
	    }
	  communal_length = data[0];
	  data += 1;
	  len -= 1;
	  break;
	}

      lcomdef (name, type_index, communal_length);
    }
}

void OMFProcessor::lcomdef (PString *name, unsigned short type,
			    unsigned long length)
{
}


virtual void OMFProcessor::raw_cextdef (byte *data, unsigned short len)
{
}

virtual void OMFProcessor::raw_comdat (byte *data, unsigned short len)
{
}

virtual void OMFProcessor::raw_comdat32 (byte *data, unsigned short len)
{
}

virtual void OMFProcessor::raw_linsym (byte *data, unsigned short len)
{
}

virtual void OMFProcessor::raw_linsym32 (byte *data, unsigned short len)
{
}

virtual void OMFProcessor::raw_alias (byte *data, unsigned short len)
{
}

virtual void OMFProcessor::raw_nbkpat (byte *data, unsigned short len)
{
}

virtual void OMFProcessor::raw_nbkpat32 (byte *data, unsigned short len)
{
}

virtual void OMFProcessor::raw_llnames (byte *data, unsigned short len)
{
}

virtual void OMFProcessor::raw_vernum (byte *data, unsigned short len)
{
}

virtual void OMFProcessor::raw_vendext (byte *data, unsigned short len)
{
}

virtual void OMFProcessor::raw_library_header_record (byte *data,
						      unsigned short len)
{
}

virtual void OMFProcessor::raw_library_end_record (byte *data,
						   unsigned short len)
{
}

